@@ -63,7 +63,8 @@ install () {
         mkdir -p $SNAKEPIT
         mkdir -p $BASE/versions
         update_pyenv $VERSION
-        # -Ofast makes the build take too long and times out Travis.
+        # -Ofast makes the build take too long and times out Travis. It also affects
+        # process-wide floating-point flags - see: https://github.com/gevent/gevent/pull/1864
         CFLAGS="-O1 -pipe -march=native" $BASE/pyenv/plugins/python-build/bin/python-build $VERSION $DESTINATION
     fi
 
@@ -120,7 +121,7 @@ for var in "$@"; do
             install 3.8.2 python3.8 3.8.d
             ;;
         3.9)
-            install 3.9-dev python3.9 3.9.d
+            install 3.9.0 python3.9 3.9.d
             ;;
         pypy2.7)
             install pypy2.7-7.3.1 pypy2.7 pypy2.7.d
