@@ -34,6 +34,7 @@ from gevent.testing.sysinfo import RESOLVER_DNSPYTHON
 from gevent.testing.sysinfo import RESOLVER_ARES
 from gevent.testing.sysinfo import PY2
 from gevent.testing.sysinfo import PYPY
+
 import gevent.testing.timing
 
 
@@ -45,7 +46,8 @@ RUN_ALL_HOST_TESTS = os.getenv('GEVENTTEST_RUN_ALL_ETC_HOST_TESTS', '')
 
 
 def add(klass, hostname, name=None,
-        skip=None, skip_reason=None):
+        skip=None, skip_reason=None,
+        require_equal_errors=True):
 
     call = callable(hostname)
 
@@ -64,33 +66,39 @@ def add(klass, hostname, name=None,
 
     def test_getaddrinfo_http(self):
         x = hostname() if call else hostname
-        self._test('getaddrinfo', x, 'http')
+        self._test('getaddrinfo', x, 'http',
+                   require_equal_errors=require_equal_errors)
     test_getaddrinfo_http.__name__ = 'test_%s_getaddrinfo_http' % name
     _setattr(klass, test_getaddrinfo_http.__name__, test_getaddrinfo_http)
 
     def test_gethostbyname(self):
         x = hostname() if call else hostname
-        ipaddr = self._test('gethostbyname', x)
+        ipaddr = self._test('gethostbyname', x,
+                            require_equal_errors=require_equal_errors)
         if not isinstance(ipaddr, Exception):
-            self._test('gethostbyaddr', ipaddr)
+            self._test('gethostbyaddr', ipaddr,
+                       require_equal_errors=require_equal_errors)
     test_gethostbyname.__name__ = 'test_%s_gethostbyname' % name
     _setattr(klass, test_gethostbyname.__name__, test_gethostbyname)
 
-    def test3(self):
+    def test_gethostbyname_ex(self):
         x = hostname() if call else hostname
-        self._test('gethostbyname_ex', x)
-    test3.__name__ = 'test_%s_gethostbyname_ex' % name
-    _setattr(klass, test3.__name__, test3)
+        self._test('gethostbyname_ex', x,
+                   require_equal_errors=require_equal_errors)
+    test_gethostbyname_ex.__name__ = 'test_%s_gethostbyname_ex' % name
+    _setattr(klass, test_gethostbyname_ex.__name__, test_gethostbyname_ex)
 
     def test4(self):
         x = hostname() if call else hostname
-        self._test('gethostbyaddr', x)
+        self._test('gethostbyaddr', x,
+                   require_equal_errors=require_equal_errors)
     test4.__name__ = 'test_%s_gethostbyaddr' % name
     _setattr(klass, test4.__name__, test4)
 
     def test5(self):
         x = hostname() if call else hostname
-        self._test('getnameinfo', (x, 80), 0)
+        self._test('getnameinfo', (x, 80), 0,
+                   require_equal_errors=require_equal_errors)
     test5.__name__ = 'test_%s_getnameinfo' % name
     _setattr(klass, test5.__name__, test5)
 
@@ -187,17 +195,20 @@ class TestCase(greentest.TestCase):
             return type(result1) is not type(result2)
         return repr(result1) != repr(result2)
 
-    def _test(self, func_name, *args):
+    def _test(self, func_name, *args, **kwargs):
         """
         Runs the function *func_name* with *args* and compares gevent and the system.
 
+        Keyword arguments are passed to the function itself; variable args are
+        used for the socket function.
+
         Returns the gevent result.
         """
         gevent_func = getattr(gevent_socket, func_name)
         real_func = monkey.get_original('socket', func_name)
 
         tester = getattr(self, '_run_test_' + func_name, self._run_test_generic)
-        result = tester(func_name, real_func, gevent_func, args)
+        result = tester(func_name, real_func, gevent_func, args, **kwargs)
         _real_result, time_real, gevent_result, time_gevent = result
 
         if self.verbose_dns and time_gevent > time_real + 0.02 and time_gevent > 0.03:
@@ -213,14 +224,17 @@ class TestCase(greentest.TestCase):
 
         return gevent_result
 
-    def _run_test_generic(self, func_name, real_func, gevent_func, func_args):
+    def _run_test_generic(self, func_name, real_func, gevent_func, func_args,
+                          require_equal_errors=True):
         real_result, time_real = self.run_resolver(real_func, func_args)
         gevent_result, time_gevent = self.run_resolver(gevent_func, func_args)
         if util.QUIET and self.should_log_results(real_result, gevent_result):
             util.log('')
             self.__trace_call(real_result, time_real, real_func, func_args)
             self.__trace_call(gevent_result, time_gevent, gevent_func, func_args)
-        self.assertEqualResults(real_result, gevent_result, func_name)
+
+        self.assertEqualResults(real_result, gevent_result, func_name,
+                                require_equal_errors=require_equal_errors)
         return real_result, time_real, gevent_result, time_gevent
 
     def _normalize_result(self, result, func_name):
@@ -383,6 +397,12 @@ class TestCase(greentest.TestCase):
         # one resolver than we do with the other resolver.
         # So as long as they have some subset in common,
         # we'll take it.
+        errors = isinstance(real_result, Exception) + isinstance(gevent_result, Exception)
+        if errors == 2:
+            return True
+        if errors == 1:
+            return False
+
         if not set(real_result).isdisjoint(set(gevent_result)):
             return True
         return self._generic_compare_results(real_result, gevent_result, func_name)
@@ -411,7 +431,8 @@ class TestCase(greentest.TestCase):
         # As for getaddrinfo, we'll just check the ipaddrlist has something in common.
         return not set(real_result[2]).isdisjoint(set(gevent_result[2]))
 
-    def assertEqualResults(self, real_result, gevent_result, func_name):
+    def assertEqualResults(self, real_result, gevent_result, func_name,
+                           require_equal_errors=True):
         errors = (
             OverflowError,
             TypeError,
@@ -421,7 +442,8 @@ class TestCase(greentest.TestCase):
             socket.herror,
         )
         if isinstance(real_result, errors) and isinstance(gevent_result, errors):
-            self._compare_exceptions(real_result, gevent_result, func_name)
+            if require_equal_errors:
+                self._compare_exceptions(real_result, gevent_result, func_name)
             return
 
         real_result = self._normalize_result(real_result, func_name)
@@ -635,21 +657,22 @@ class TestGeventOrg(TestCase):
     # trying www.gevent.org which is a CNAME to readthedocs.org then worked, but it became
     # an alias for python-gevent.readthedocs.org, which is an alias for readthedocs.io,
     # and which also has multiple addresses. So we run the resolver twice to try to get
-    # the different answers, if needed.
+    # the different answers, if needed. Even then it's not enough, so
+    # we normalize the two addresses we get to a single one.
     HOSTNAME = 'www.gevent.org'
 
 
-    if RESOLVER_NOT_SYSTEM:
-        def _normalize_result_gethostbyname(self, result):
-            if result == '104.17.33.82':
-                result = '104.17.32.82'
-            return result
 
-        def _normalize_result_gethostbyname_ex(self, result):
-            result = super(TestGeventOrg, self)._normalize_result_gethostbyname_ex(result)
-            if result[0] == 'python-gevent.readthedocs.org':
-                result = ('readthedocs.io', ) + result[1:]
-            return result
+    def _normalize_result_gethostbyname(self, result):
+        if result == '104.17.33.82':
+            result = '104.17.32.82'
+        return result
+
+    def _normalize_result_gethostbyname_ex(self, result):
+        result = super(TestGeventOrg, self)._normalize_result_gethostbyname_ex(result)
+        if result[0] == 'python-gevent.readthedocs.org':
+            result = ('readthedocs.io', ) + result[1:]
+        return result
 
     def test_AI_CANONNAME(self):
         # Not all systems support AI_CANONNAME; notably tha manylinux
@@ -769,10 +792,22 @@ class TestInternational(TestCase):
 # the 2008 version of idna encoding, whereas on Python 2,
 # with the default resolver, it tries to encode to ascii and
 # raises a UnicodeEncodeError. So we get different results.
+
+# Starting 20221027, on GitHub Actions and *some* versions of Python,
+# we started getting a different error result from our own resolver
+# compared to the system. This is very weird because our own resolver
+# calls the system. I can't reproduce locally. Perhaps the two
+# different answers are because of caching? One from the real DNS
+# server, one from the local resolver library? Hence
+# require_equal_errors=False
+# ('system:', "herror(2, 'Host name lookup failure')",
+#  'gevent:', "herror(1, 'Unknown host')")
 add(TestInternational, u'президент.рф', 'russian',
     skip=(PY2 and RESOLVER_DNSPYTHON),
-    skip_reason="dnspython can actually resolve these")
-add(TestInternational, u'президент.рф'.encode('idna'), 'idna')
+    skip_reason="dnspython can actually resolve these",
+    require_equal_errors=False)
+add(TestInternational, u'президент.рф'.encode('idna'), 'idna',
+    require_equal_errors=False)
 
 @skipWithoutExternalNetwork("Tries to resolve and compare hostnames/addrinfo")
 class TestInterrupted_gethostbyname(gevent.testing.timing.AbstractGenericWaitTestCase):
